# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2016 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/damona/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################


import pathlib
import glob
import os
import yaml
from yaml import Loader
from damona import logger


class Registry():
    """

    The registry contains a dictionary with names of images and metadata
    information about them::

        r = Registry()
        r.registry
        r.registry['prokka_1.14.5']['download']

    """

    def __init__(self, from_url=None):
        if from_url == "damona":
            from_url = "https://biomics.pasteur.fr/drylab/damona/registry.txt"
        elif from_url:
            assert from_url.startswith('http')
            assert from_url.endswith('registry.txt')
        self.from_url = from_url
        self.registry = {}
        self.discovery()

    def _read_registry(self, registry):
        if os.path.exists(registry) is False: #pragma: no cover
            raise IOError("incorrect input filename {}".format(registry))

        # read the yaml
        data = yaml.load(open(registry, "r").read(), Loader=Loader)

        # some checks
        for k, v in data.items():
            if "class" not in v: #pragma: no cover
                logger.warning("missing class in {}".format(registry))
            if "version" not in v: #pragma: no cover
                logger.warning("missing version in {}".format(registry))
            if "download" not in v: #pragma: no cover
                logger.warning("missing download in {}".format(registry))
            if data[k]["class"] == "exe":
                if "binaries" not in v: #pragma: no cover
                    logger.warning("missing binaries field in {}".format(registry))
        return data

    def discovery(self):

        if self.from_url:
            self._url_discovery()
        else:
            self._local_discovery()

    def _url_discovery(self):
        self.registry = {}
        import  urllib.request

        response = urllib.request.urlopen(self.from_url)
        html = response.read()
        html = html.decode('utf-8')
        _class = None
        for registry in html.split("\n"):
            if registry.startswith("["):
                _class = registry.replace('[', '').replace(']','')
                continue
            if registry.strip() == "":
                continue
            else:
                data = {}
                name,version = registry.split("_")
                version = version.rsplit(".", 1)[0]
                # todo for now only exe fills the binaries assuming name of
                # package is the name of the binary
                self.registry[name+"_"+version] = {
                    "name": name+"_"+version,   
                    "download": self.from_url.replace("registry.txt", registry),
                    "version": version,
                    "class": _class, 
                    'binaries': [{name:name}]}

    def _local_discovery(self):

        from damona.recipes import __path__
        self._registry_files = glob.glob(__path__[0] + '/*/registry.yaml')
        self._singularity_files = glob.glob(__path__[0] + '/*/Singularity.*')

        self.registry = {}

        for registry in self._registry_files:
            data = self._read_registry(registry)
            for k, v in data.items():
                name = k.replace("Singularity.", "").lower()
                if name not in self.registry:
                    self.registry[name] = v
                else: #pragma: no cover
                    raise ValueError("found a duplicated name {}".format(name))

        for filename in self._singularity_files:
            p = pathlib.Path(filename)
            if (p.parent / "registry.yaml").exists() is False: #pragma: no cover
                logger.warning("Missing registry in {}. You may use 'damona registry {} to start with".format(p.parent, p.parent))

    def create_registry(self, path):
        #recipes = glob.glob(pathlib.Path(path).absolute() / "Singularity*")
        recipes = glob.glob(path +  "/Singularity*")
        for recipe in recipes:
            try:version = recipe.rsplit("_")[-1]
            except: version = "fillme"
            print("""{}:
    version: {}
    class: choose one in : exe, env,set
    download: library://path_to_image
    binaries: if env delete this file otherwise complete with list of binaries
        - name1 in singularity: name1 in your env
        - name1 in singularity: name1 in your env
    """.format(recipe.rsplit("/",1)[1], version))

    def get_list(self, pattern=None):
        if self.from_url is None:
            from damona.recipes import __path__
            recipes = glob.glob(__path__[0] + '/*/Singularity.*')
            recipes = [os.path.basename(x) for x in recipes]
            recipes = [x.replace("Singularity.", "").lower() for x in recipes]
            recipes = [x.replace("_", ":").lower() for x in recipes]
            recipes = sorted(recipes)

            if pattern:
                recipes = [x for x in recipes if pattern in x]

            return recipes
        else:
            self._url_discovery()
            names = []
            for k,v in self.registry.items():
                names.append(k.replace("_", ":" ))
            if pattern:
                names = [x for x in names if pattern in x]
            return names



# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2016 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
import os
import sys
import packaging.version

from spython.main import Client
from damona import images_directory
from damona import Registry
from damona import logger

logger.level = 10


class Pull():
    """Manager to download container images


    """
    def __init__(self, dryrun=False, from_url=None):
        self.registry = Registry(from_url=from_url).registry
        self.from_url = from_url
        self.dryrun = dryrun

    def pull(self, name, output_name=None, pull_folder=images_directory,
        force=False):

        # the name must be valid. we use _ to separate name and version for the
        # singularity filemane but the image have a tag (version) and name
        # separated by a :

        # Our registry uses underscore. The output filename should use _ as well
        # However, for users and image stored on sylabs or singularity hub, we
        # use the character :
        #
        if ":" in name:
            logger.info(f"Looking for {name}...")
            # e.G. fastqc:0.11.9
            registry_name = name.replace(":", "_")
            if registry_name not in self.registry.keys():
                logger.critical("invalid image name provided: {}. Choose amongst {}".format(
                    name, self.registry.keys()))
                sys.exit(1)
            else:
                logger.info(f"... found. Please wait while downloading (if not already done)")
        else:
            # e.g. fastqc
            logger.info(f"No tag found after {name}. We will download the latest version")
            registry_name = name
            # we look only at the prefix name, not the tag. so we should get the
            # registry names from self.registry that have the prefix in common, 
            # then extract the versions, and figure out the most recent.
            candidates = [x for x in self.registry.keys() if x.split("_")[0] == registry_name]
            if len(candidates) == 0:
                logger.critical(f"No image found for {name}. MAke sure it is correct using 'damona list' command")
                sys.exit(1)
            names = [x.split("_")[0] for x in candidates]
            versions = [x.split("_")[1] for x in candidates]

            version = max([packaging.version.parse(ver) for ver in versions])
            name = names[0] 
            registry_name = name + '_' + str(version)
            logger.info("pulling {}".format(registry_name))


        download_name = self.registry[registry_name]['download']

        if output_name is None:
            output_name = registry_name + ".img"

        if self.dryrun: 
            pass
        else: # pragma: no cover
            # By default it downlaods from syslab if it can be found, but one
            # can provide an url
            if self.from_url:
                # The client does not support external https link other than
                # dockker, library, shub. 
                cmd = f"singularity pull  --dir {pull_folder} "
                if force:
                    cmd += " --force "
                cmd += f"{download_name}"
                import subprocess

                subprocess.call(cmd.split())
            else:
                Client.pull(str(download_name), name=output_name, 
                    pull_folder=pull_folder,
                    force=force) 
            logger.info("File {} uploaded to {}".format(name, images_directory))

        # Now, create an alias
        path = images_directory

        _class = self.registry[registry_name]['class']
        if _class == "exe":
            # one binary name. we use the first one
            cmd = """singularity run {}/{} ${{1+"$@"}} """
            cmd = cmd.format(images_directory, output_name)
            from damona import Environ
            env = Environ()
            bin_directory = env.get_current_env() + "/bin"
            bin_name = bin_directory +"/"+  registry_name.split("_")[0]
            if self.dryrun: #pragma: no cover
                pass
            else: #pragma: no cover
                with open(bin_name, "w") as fout:
                    fout.write(cmd)
                os.chmod(bin_name, 0o755)
                logger.info("Creating binary {}".format(bin_name))
        elif _class == "env": #pragma: no cover
            pass
        elif _class == "set": # pragma: no cover
            raise ImplementedError
        else: # pragma: no cover
            raise ValueError





from damona import script
from damona import pull
import subprocess
import builtins

def test_damona_app():
    from click.testing import CliRunner
    from damona.script import install, build, develop, env, activate
    from damona.script import list as _list
    runner = CliRunner()

    # isntall
    results = runner.invoke(install, ['fastqc:0.11.9', '--dryrun'])
    assert results.exit_code == 0

    results = runner.invoke(_list)
    assert results.exit_code == 0

    import damona
    results = runner.invoke(develop, ["--path", damona.__path__[0] + "/recipes/fastqc"])
    assert results.exit_code == 0
    
    import damona
    results = runner.invoke(env, [])
    # wrong
    results = runner.invoke(env, ["--create", ".dummy_test", "--delete", ".dummy_test"])
    # good
    results = runner.invoke(env, ["--create", ".dummy_test"])
    assert results.exit_code == 0
    results = runner.invoke(env, ["--create", ".dummy_test"])
    assert results.exit_code == 0
    results = runner.invoke(activate, [".dummy_test"])
    assert results.exit_code == 0

    import mock
    with mock.patch.object(builtins, 'input', lambda _: 'y'):
        results = runner.invoke(env, ["--delete", ".dummy_test"])
        assert results.exit_code == 0

def test_environ():
    from damona.environ import Environ
    env = Environ()
    env.N
    env.environments
    env.create(".dummy_test")
    import mock
    with mock.patch.object(builtins, 'input', lambda _: 'y'):
        env.delete(".dummy_test")


def test_python_pull():
    p = pull.Pull(dryrun=True)
    p.pull("fastqc:0.11.9")
    p.pull("fastqc") # latest
    try:
        p.pull("fastqc1")
        assert False
    except:
        assert True

def test_python_registry():
    from damona import registry
    r = registry.Registry()
    r.get_list()

    from damona.recipes import __path__
    path = __path__[0] + "/fastqc"
    r.create_registry(path)

    r = registry.Registry("damona")
    r.get_list()


def test_help():
    cmd = "damona --help"
    subprocess.call(cmd.split())
    cmd = "damona pull --help"
    subprocess.call(cmd.split())


def test_list():
    cmd = "damona list --help"
    subprocess.call(cmd.split())

    cmd = "damona list --pattern qc"
    subprocess.call(cmd.split())

def test_develop():
    try:
        cmd = "damona develop"
        subprocess.call(cmd.split())
        assert False
    except:
        assert True

    from damona.recipes import __path__
    cmd = "damona develop --path {}".format(__path__[0] + "/fastq"  )
    subprocess.call(cmd.split())

def test_pull_dryrun():
    cmd = "damona pull fastqc:0.11.9 --dryrun"  
    subprocess.call(cmd.split())
    
def test_pull_wrong():
    # non existing image
    cmd = "damona pull tartuffe"  
    try:
        subprocess.call(cmd.split())    
        assert False
    except:
        assert True



# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2016 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
from damona import env_directory
import os
import shutil
from damona import logger


class Environ():
    def __init__(self):
        pass

    @staticmethod
    def get_current_env():
        if 'DAMONA_ENV' not in os.environ:
            from damona import damona_config_path
            return damona_config_path
        else:
            return os.environ['DAMONA_ENV'] 
    
    def _get_N(self):
        return len(self.environments)
    N = property(_get_N)

    def _get_envs(self):
        envs = os.listdir(str(env_directory))
        return envs
    environments = property(_get_envs)

    def delete(self, env_name):
        env_path = env_directory / env_name
        if os.path.exists(env_path) is False:
            logger.error("{} does not exists".format(env_path))
        else:
            try:
                os.rmdir(env_path)
            except OSError:
                logger.warning("Will delete all contents of {}. Although this concerns only aliases you will lose your environement".format(env_path))
                ret = input("Are you sure you want to proceed ? (N/y)")
                if ret == 'y':
                    shutil.rmtree(env_path)

    def activate(self, env_name):
        assert env_name in self.environments, "invalid name. use 'damona env' to get the list"
        env_path = env_directory / env_name
        # this cannot be done permanently from python
        print("This feature is not fully implemented yet. Please set the environmental "
              " variable manually. For information about environmental variable, see "
               "e.g. https://www.schrodinger.com/kb/1842")
        print("Under bash type:\n")
        print('    export DAMONA_ENV="{}"\n'.format(env_path))

    def create(self, env_name):
        env_path = env_directory / env_name
        if os.path.exists(env_path):
            logger.warning("{} exists already".format(env_path))
        else:
            try:
                os.mkdir(env_path)
                os.mkdir(env_path / "bin")
                logger.info("Created {} in {}".format(env_path, env_directory))
            except:
                logger.warning("Something went wrong. Could not create {}".format(env_path))




# -*- coding: utf-8 -*-

###########################################################################
# Damona is a project to provide reproducible containers                  #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2020  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# Damona is free software: you can redistribute it and/or modify          #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# Damona  is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""".. rubric:: Standalone application dedicated to conversion"""
import os
import glob
import json
import sys
import colorlog
import textwrap
import subprocess
import click
from damona  import version,  images_directory
import pathlib

__all__ = ["main"]

from damona import logger
logger.level = 10


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])



@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version)
def main():
    """This is the main help"""
    pass


@main.command()
@click.argument('name', type=click.STRING)
@click.option('--force', is_flag=True)
@click.option('--dryrun', is_flag=True, 
    help="If set, do not pull the image and do not create the binary alias")
@click.option('--from-url', help="""download image from a remote URL. The URL must
contain a registry.txt as explained on damona.readthedocs.io""") 
@click.option('--output-directory', default=images_directory, 
    help="""Where to save the image (default: {})""".format(images_directory))
def install(**kwargs):
    """Download and install an  image given its name and version (optional)

    Under Linux the DAMONA PATH is under ~/.config/damona

    """
    name = kwargs['name']
    from damona.pull import Pull
    p = Pull(dryrun=kwargs['dryrun'], from_url=kwargs['from_url'])
    p.pull(name, pull_folder=kwargs["output_directory"],
        force=kwargs['force'])


@main.command()
@click.argument('filename', required=True, type=click.STRING)
@click.option('--output-name', default=None, 
    help="default to the singularity extension and tag")
def build(**kwargs): #pragma: no cover
    """Build a container from a local recipes or damona recipes.

    You must have sudo permissions
    If FILENAME is not local, try to find it in Damona and build it.

        # a local recipes
        damona build ./Singularity.recipe

       # a recipe to be found in damona
        damona build salmon:1.3.0

    """
    cmd = "sudo singularity build {} {}"
    output_name = kwargs.get('output_name')
    if output_name is None:
        # get filename and save locally
        output_name = str(pathlib.Path(kwargs['filename']).name)
        output_name = output_name.replace("Singularity.", "")
        output_name = output_name.replace(":", "_") + ".img"
    else:
        assert output_name.endswith(('.sif', 'img'))

    if os.path.exists(kwargs['filename']) and os.path.isdir(kwargs['filename']) is False:
        cmd = cmd.format(output_name, kwargs['filename'])
    else:
        from damona import registry
        reg = registry.Registry()
        user_name = kwargs['filename']
        candidate = [x for x in reg._singularity_files 
            if x.endswith(user_name.replace(':', '_', 1))]
        if len(candidate) == 0:
            logger.critical(f'unknown image name {user_name}. use the command' +
                ' "damona list" to get correct names')
            sys.exit(1)
        logger.info("Building using damona recipes for {}".format(candidate[0]))
        cmd = cmd.format(output_name, candidate[0].replace(':', '_'))
    subprocess.call(cmd.split())


@main.command()
@click.option('--pattern', default=None, 
    help="restrict the output list keeping those with this pattern")
@click.option('--from-url', help="""download registry from remote URL. The URL must
contain a registry.txt as explained on damona.readthedocs.io. If you just type
'damona', the URL will be biomics.pasteur.fr/drylab/damona/registry.txt""") 
def list(**kwargs):
    """List all available images from Damona

    damona list
    damona list --from-url https://biomics.pasteur.fr/drylab/damona/registry.txt
    #This is the same command as above (alias)
    damona list --from-url damona

    """
    from damona.registry import Registry
    modules = Registry(from_url=kwargs["from_url"]).get_list(pattern=kwargs['pattern'])
    print(", ".join(modules))


@main.command()
@click.option('--create', type=click.STRING) 
@click.option('--delete', type=click.STRING) 
def env(**kwargs):
    """List the Damona environments"""
    from damona import Environ
    envs = Environ()
    if kwargs['create'] is None and kwargs['delete'] is None:
        print(f"There is currently only one base environment and {envs.N} user environment.")
        if envs.N !=0:
            print("Here are the current environment(s): ")
            for this in envs.environments:
                print(" -  {}".format(this))
        current_env = envs.get_current_env()
        logger.info("""Your current env is {}.

You can overwrite this behaviour by setting DAMONA_ENV variable.
For example under bash:

    export DAMONA_ENV="~/.config/damona/envs/your_best_env"


        """.format(current_env))
    elif kwargs['create'] and kwargs['delete']: # mutually exclusive
        logger.error("you cannot use --delete and --create together")
    elif kwargs['delete']:
        envs.delete(kwargs['delete'])
    elif kwargs['create']:
        envs.create(kwargs["create"])

@main.command()
@click.argument('name', required=True, type=click.STRING)
def activate(**kwargs):
    """activate a damona environment
    
    List the environments

        damona env

    and activate a valid one:

        damona activate example

    """
    from damona import Environ
    env = Environ()
    env.activate(kwargs['name'])

@main.command()
@click.option('--path', required=True, 
    help="path to recipes directory where Singularity file(s) can be found")
def develop(**kwargs):
    """Developers kit (eg build registry)"""
    from damona.registry import Registry

    if kwargs['path'] and os.path.exists(kwargs['path']):
        modules = Registry().create_registry(kwargs['path'])
    else:
        logger.critical('Please provide a valid path where to find Singulary from damona software')

if __name__ == "__main__": #pragma: no cover
    main()



import pkg_resources
import os
import pathlib

try:
    version = pkg_resources.require("damona")[0].version
except:   #pragma: no cover
    version = ">=0.8.3"


import colorlog

handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(
	'%(log_color)s%(levelname)s:%(name)s: %(message)s'))
logger = colorlog.getLogger("damona")
logger.addHandler(handler)

from easydev import CustomConfig
configuration = CustomConfig("damona", verbose=True)
damona_config_path = configuration.user_config_dir


_damona_path = pathlib.Path(damona_config_path)


# First the env directory then, the sub-directories. 
try:
    env_directory = _damona_path / 'envs'
    env_directory.mkdir()
except: # pragma: no cover
    pass # exists already

try:
    _bin_directory = _damona_path / 'bin'
    _bin_directory.mkdir()
except: # pragma: no cover
    pass # exists already

try:
    images_directory = _damona_path / 'images'
    images_directory.mkdir()
except: #pragma: no cover
    pass # exists already

from damona.registry import Registry
from damona.environ import Environ
